--*******************************************************************************************************************************
--
-- Description: This module collects data produced from the DMA write transaction and saves data into a dma buffer. 
--				These functions emulate the dma functions controlled by the PLDA EZDMA2 core in the SigFPGA sample project.
--					
-- Rev History: 8-22-2014:	Added functionality to control read channels allowing only one channel to use the bus at a time.
--				4-30-2014:	Corrected the state machine, which was dropping the first sample and saving the last sample twice to the dma buffer.
--				3-31-2014:	Initial release.		  
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

use work.model_specific_package.all;
use work.transceiver_package.all;
use work.tb_package.all;

entity rx_dma_emulator is
port (		
    -- Clock and reset.
    user_clk		: in std_logic;
    user_reset		: in std_logic;
    
	-- Direct memory access (dma) interface connections to the application logic.
	dma_rd_ch 		: out std_logic;							-- Logic one indicates that the core is consuming data from a DMA write request.
	dma_rdchannel 	: out std_logic;							-- Logic one indicates that data is needed on this DMA channel.
	dma_status		: out std_logic_vector(3 downto 0);			-- Status of the DMA.
	dma_ch_go		: in std_logic;								-- DMA control settings from dma_on_demand core.
	dma_rddata		: in std_logic_vector(AXI_WIDTH-1 downto 0)	-- DMA local data bus used for write transactions.
	);
end rx_dma_emulator;   

architecture structural of rx_dma_emulator is
	-- The size of the array is burst_count by burst_size. 
	type RX_DATA is array (1 to (BURST_SIZE/(AXI_WIDTH/8))*BURST_COUNT) of std_logic_vector(AXI_WIDTH-1 downto 0); -- Array containing AXI_WIDTH-bit DMA buffer samples. The burst size has been divided by 16 because there are 16 bursts per line.
	type RX_CHNL_DATA is array (RX_CHANNEL_COUNT downto 1) of RX_DATA; -- Array for each RX channel.
	        
    -- Data signals
	signal rddata	  : RX_DATA;	-- Data bus which collects all burst data recieved during a write transaction. 
	
	-- Control Signals
	signal j		: integer;		-- Counter to control which address data is being written to. 
	signal k		: integer; 		-- Counter to control the burst currently being written.
	signal dma_ch_go_R	: std_logic; -- Signal is registered to account for read latency. 
    
    type pciestate is (pcie_idle, pcie_waiting, pcie_working, pcie_finish_collection, pcie_burst_complete);	-- Type definitions for the PCI Express state machine.
	signal pcie_state		: pciestate;  					-- Current PCI Express machine state.

                         
begin                                                          

	--************************************************* STATE MACHINE TRANSITIONS **************************************************
	PCI_EXPRESS_STATE_MACHINE : process (user_clk, user_reset)
	begin
		if user_reset = '1' then
		   	pcie_state <= pcie_idle;
		   	
		elsif rising_edge(user_clk) then
			dma_ch_go_R 	<= dma_ch_go;
			
			case pcie_state is
				
				-- Wait for dma_ch_go to be set to start dma transfer. 
				when pcie_idle =>
					if dma_ch_go /= '0' and k <= BURST_COUNT then
						pcie_state <= pcie_waiting;
					end if;
					
				-- Wait for a 2 clock latency to synchronize with the FWFT FIFO. 
				when pcie_waiting => 
					if dma_ch_go /= '0' then
						pcie_state <= pcie_working;
					end if;
	

				-- Save data into the dma buffer while a dma transfer completes.
				when pcie_working =>
					if j = (BURST_SIZE/(AXI_WIDTH/8))*k - 1 then
						pcie_state <= pcie_finish_collection;
					else 
						pcie_state <= pcie_working;						
					end if;
					
				-- Finish collecting the last valid dma data samples transmitted.
				when pcie_finish_collection => 
					if j = (BURST_SIZE/(AXI_WIDTH/8))*k then
						pcie_state <= pcie_burst_complete;
					else 
						pcie_state <= pcie_finish_collection;						
					end if;					
				
				-- Go back to idle after the dma is complete.
				when pcie_burst_complete =>
					pcie_state <= pcie_idle;
					
				-- Define the default state to be dma_idle.
				when others => pcie_state <= pcie_idle;
			end case;
		end if;
	end process;
	
	--*************************************************** STATE MACHINE EVENTS *****************************************************
	process (user_clk,user_reset)
	begin
		if user_reset = '1' then
			j <= 1;					
			k <= 1;					
			dma_rdchannel <= '0';	                                          
			dma_rd_ch <= '0';		   
		   	dma_status <= (others => '0');
		   	
		elsif rising_edge(user_clk) then			
			if pcie_state = pcie_idle then 
				dma_status <= x"0";
				dma_rdchannel <= '0';	                                         
				dma_rd_ch <= '0';	 
			end if; 
		
			-- Set control signals high only after a request is received and until all data is logged.
			if pcie_state = pcie_waiting then 
				dma_status <= x"8";		-- Set dma_status high while channel is collecting data from a burst.
				dma_rdchannel <= '1';	-- Set dma_rdchannel high after a dma read request is recieved to grant data transmission.                                          
				dma_rd_ch <= '1';		-- Send read enable to dma_write_fifo.   
			end if; 

			-- Save incoming dma buffer data to memory array.
			if pcie_state = pcie_working then
				rddata(j) <= dma_rddata;
				j <= j + 1;
				dma_status <= x"8";
				dma_rdchannel <= '1';	                                         
				dma_rd_ch <= '1';	   
			end if; 
	
			-- Save the last two valid read data samples to memory array.
			if pcie_state = pcie_finish_collection then
				rddata(j) <= dma_rddata;
				j <= j + 1;
				dma_status <= x"8";		-- Keep dma_status high while dma buffer data is still being collected. 
				dma_rdchannel <= '0';	-- Turn off dma_rdchannel after a full burst completes.                                         
				dma_rd_ch <= '0';		-- Turn off read enable after a full burst completes   
			end if; 
			
			-- Reset all control signals back to zero until another burst is initiated.
			if pcie_state = pcie_burst_complete then
				dma_status <= x"0"; 		
				dma_rdchannel <= '0';	                                         
				dma_rd_ch <= '0';	
				-- Look to see to see if all the bursts have occurred, filling the dma buffer with all the requested data; if not then progress to the next burst. 
				if j >= BURST_SIZE/16 and k >= BURST_COUNT then
					j <= 1;
					k <= 1;
				else 
					k <= k + 1;
				end if;
			end if;
		end if;
	end process; 
	          	
end structural;

